/****************************************************************************
    reader.cpp  25/Jun/2000
    Copyright (C)  Ernesto F. Artaza  <artaza@info.isctn.edu.cu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
****************************************************************************/

#include "reader.h"
#include <string.h>
#include <stdio.h>
#include <qlist.h>
#include <stdarg.h>
#include <stdlib.h>
#include <qobject.h>
#include <qapplication.h>

#define STREQU(s1, s2) ((s1==s2) || (s1 && s2 && (strcmp(s1, s2)==0)))

void error(const char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	printf("vtk error: ");
	vprintf(fmt, args);
	va_end(args);
	exit(1);
}

QObject *FindChild(QObject *Top, const char *Name)
{
	if (Name && Top->name() && (strcmp(Top->name(), Name)==0)) return Top; 

	QListT<QObject> *list = (QListT<QObject>*) Top->children();
	if (list)
	for ( list->first(); list->current(); list->next() )
	if ((Top = FindChild(list->current(), Name))) return Top;
	return NULL;
}

void  ReadPrivateObject(TReader &r, QObject *o, void *Reader, ClassInfo *classinfo)
{
	o->setName(r.readString());
	QGetPosEvent	event(2);
	event.X = r.readInteger();
	event.Y = r.readInteger();
	qApp->sendEvent(o, &event);
	((ReaderProc) Reader)(r, o, classinfo);
}

QObject	*TReader::readObject(QObject *o)
{	
	Object = o;
	if ( Root == NULL ) Root = o;

	ClassInfo *classinfo = ClassInfo::Ancestor( o );

	if (o && classinfo) 
	{
		void *Reader = readers.mapper(classinfo->name, "");
		if (Reader) ReadPrivateObject(*this, o, Reader, classinfo);
		else
		{
			readProperties(o, classinfo);
			readSignals(o);
			readChildren(o);
		}
	} else o = NULL;
	
	return o;
}

void  TReader::readProperties(QObject *o, ClassInfo *classinfo)
{	
	for (;;) {
		PropPath = readPropPath();
		if (PropPath.length()==0) return;
		readProperty(o, classinfo, PropPath);
	}	
}


void  TReader::readProperty(QObject *o, ClassInfo *classinfo, QString Name)
{
	char	tmp[256], *p;
	PropInfo *propinfo;

	strcpy(tmp, Name);


	do {	
		p = strstr(tmp, ".");
		if (!p) propinfo = classinfo->Prop(tmp); else
		{
			*p= 0;
			propinfo = classinfo->Prop(tmp);
			if (!propinfo || (propinfo->type->kind != tkclass))
			{	
				// error reading not class subproperty
				propinfo = NULL;
				break;
			}
			classinfo = (ClassInfo*) propinfo->type;
			o = (QObject*) getOrdProp(o, propinfo);
			if (!o) 
			{
				// error reading subproperty of NULL instance 
				break;
			}
			strcpy(tmp, p+1);
		}
	} while (p); 
	if (!propinfo) 
	{
		// error reading property %s without info
		bypassValue();	
		return;
	}

	void* Reader = readers.mapper(classinfo->name, propinfo->name);
	if (Reader) 
	{
		((ReaderProc) Reader)(*this, o, propinfo);
		return;
	}

	if (propinfo->type->kind == tkclass) Reader = readers.mapper(propinfo->type->name, "");
	if (Reader) 
	{
		((ReaderProc) Reader)(*this, o, propinfo);
		return;
	}
	
	
	switch (propinfo->type->kind) {
	case tkset:
	case tkord:
	case tkenum: 
		{ int v = readInteger();
		setOrdProp(o, propinfo, v);
		} break;
	case tkqstring:	
	case tkstring: { QString v = readString();
		setStrProp(o, propinfo, v);
		} break;
	case tkclass: {
		if (((ClassInfo*) propinfo->type)->inherits("QObject") )
		{
			QString name = readString();
//			setOrdProp(o, propinfo, FindChild( tope, name) );
		} else error("Reading wrong QObject class property %s", Name.data() ); 
		} break;			 
	default:
		error("Reading non standard property");
	}
}
	
QString TReader::readPropPath()
{
	return readString();
}

int  TReader::readInteger()
{	
        int	v = 0;
	if (readValue() != vaInt) error("Reading Integer");
	read(&v, 4);
	return v;
}


QString TReader::readString()
{
	int l = 0;
	char *v;
	QString L;
		
	if (readValue() != vaStr) error("Reading String");
	read(&l, 4);
	if (l) 
	{
		v = (char*) malloc(l+1);
		read(v, l+1); 
		L = v;
		delete v;
	}
	return  L;
}

QString TReader::readIdent()
{
  return readString();
}

TValueType TReader::readValue()
{	TValueType  v=vaNull;

	read(&v, 1);
	return v;
}

void TReader::readChildren(QObject *o)
{	
	QString className;
	do 
	{	
		className= readString();
		int	iniPos = at();
		int Size;
		
		read(&Size, 4);
		
		ClassInfo *classinfo = className.length()>0 ? ClassInfo::Ancestor(className) : NULL;
		QObject *n = (QObject*) Create(classinfo, o);

		if (n) readObject(n); 
		
		at(iniPos + 4 + Size);
		
	} while (className.length()>0);
}


void TReader::readSignals(QObject *o)
{	
	do 
	{	QString signal = readString();
		if (signal.length()==0) break;

		QString obj, member;

		obj = readString();
		member = readString();

		CheckSlot(member);

		{
			QObject *t;
			for (t=o; t; t=t->parent())
				if (STREQU(t->name(), obj)) break;

			Connect(o, signal, t, member);
		}

	} while (1);
}

void	TReader::Connect(QObject *o, QString signal, QObject *t, QString member)
{
	if (t) o->connect(o, QString("2")+signal, t, QString("1")+member);
}


QObject *ReadFormTemplate(QString File, QObject *Obj)
{
	TReader  reader(File);
	int	 Size;

	QString  className = reader.readString();
	reader.read(&Size, 4);
	
	return reader.readObject(Obj);
}

QObject *ReadFormResource(unsigned int *Size, QObject *Obj)
{
	FILE  *f = fopen("/tmp/cucurrucu", "w");
	fwrite((void*) &Size[1], *Size, 1, f);
	fclose(f);

	QObject *tmp = ReadFormTemplate("/tmp/cucurrucu", Obj);
	remove("/tmp/cucurrucu");
	return tmp;
}

QObject *TReader::Create(ClassInfo *classinfo, QObject *p)
{
	QObject *n = NULL;
	
	if (classinfo && classinfo->create)
	{
		if (!classinfo->inherits("QObject")) 
			error("Leyendo non QObject [%s]", classinfo->name);
		n = classinfo->create( p );
	} 
	return n;
}	
void TReader::bypassValue()
{
	TValueType  v = readValue();
	
	at( at() - 1 );		
	
	switch (v) {
	case vaInt: readInteger(); break;
	case vaStr: readString(); break;
	default: break;
	
/*	not implemented yet
	case vaFloat: 
	case vaList: 
	case vaBinary:
*/	
	}
}